
#include "pirat.h"
#include "piratutil.h"


/*
 (P) oint (i) n (Ra) nge (T) ree
  Copyright (C) 1996 Petr Krysl

  
  */



typedef struct pirat_pt_t  {
  void          *object;
  G3D_loc_t      loc;
}               pirat_pt_t;



struct pirat_leaf_t;

typedef struct pirat_leaf_t  {
  struct pirat_leaf_t   *descend[2];
  G3D_box_t              box;
  int                    num_of_pts;
  pirat_pt_t            *pts;
  int                    split_across;
}               pirat_leaf_t;



#define X_SPLIT 'x'
#define Y_SPLIT 'y'
#define Z_SPLIT 'z'
#define NO_SPLIT ' '



static int
split_dir (pirat_leaf_t *leaf);



static int
split_dir (pirat_leaf_t *leaf) 
{
  double _lx_ = (leaf->box.upper.x - leaf->box.lower.x);
  double _ly_ = (leaf->box.upper.y - leaf->box.lower.y);
  double _lz_ = (leaf->box.upper.z - leaf->box.lower.z);
  
  if (_lx_ > _ly_ && _lx_ > _lz_) {
    return X_SPLIT;
  }  else if (_ly_ > _lx_ && _ly_ > _lz_) {
    return Y_SPLIT;
  }  else  {
    return Z_SPLIT;
  } 
} 


static pirat_pt_t *
pirat_new_pts_array (pirat_front_end_t *pfe) 
{
  if (pfe->free_pts_array) {
    pirat_pt_t *a = pfe->free_pts_array;
    pfe->free_pts_array = NULL;
    return a;
  }  else  {
    pirat_pt_t *a;
    a = PIRAT_NEW_ARRAY (pfe->max_num_of_points_per_box,
                        pirat_pt_t);
    PIRAT_CHECK_NEW (a);
    return a;
  } 
} 

static void
pirat_free_pts_array (pirat_front_end_t *pfe, pirat_pt_t *a) 
{
  if (pfe->free_pts_array) {
    PIRAT_FREE (a);
  }  else  {
    pfe->free_pts_array = a;
  } 
} 









/*
  Create a new PiRaT (point-in-range tree) .
 */
pirat_front_end_t *
pirat_new_pirat (int max_num_of_points_per_box, double inflate_tol) 
{
  pirat_front_end_t *pirat;

  pirat = PIRAT_NEW (pirat_front_end_t);
  PIRAT_CHECK_NEW (pirat);

  pirat->num_of_inserted_points     = 0;
  pirat->max_num_of_points_per_box  = max_num_of_points_per_box;
  pirat->num_of_points_found        = 0;
  pirat->size_of_buffer             = 8 * pirat->max_num_of_points_per_box;
  pirat->entities_found             = PIRAT_NEW_ARRAY (pirat->size_of_buffer,
                                                      void *);
  pirat->inflate_tol                = inflate_tol;
  PIRAT_CHECK_NEW (pirat->entities_found);
  pirat->free_pts_array             = NULL;
  pirat->root                       = NULL;
  
  return pirat;
} 


static int
pirat_check_leaf (pirat_front_end_t *pfe, pirat_leaf_t *leaf,
                  G3D_box_t *range
                          ) ;
static void
pirat_descend_check_leaf (pirat_front_end_t *pfe,
                          pirat_leaf_t *root,
                          G3D_box_t *range
                          );

/*
   This function searches the tree for all the points which contained
   in the search range.  It returns TRUE if search was successful.
 */
int
pirat_search_range (pirat_front_end_t *pfe, G3D_box_t *range, int max_found) 
{
  pfe->num_of_points_found  = 0;
  pfe->find_max_num_objects = max_found;
  if (pfe->root) 
    pirat_descend_check_leaf (pfe, pfe->root, range);
  return TRUE;
} 


#ifdef ITERATIVE_DESCENT
static int leaf_stack_size = 0;
static pirat_leaf_t **leaf_stack = NULL;
static int leaf_stack_ptr  = 0;
#define LEAF_STACK_CLEAR() \
    {                                                                    \
     if (leaf_stack == NULL) {                                          \
       leaf_stack_size = 512,                                           \
       leaf_stack = PIRAT_NEW_ARRAY (leaf_stack_size, pirat_leaf_t *);   \
       PIRAT_CHECK_NEW (leaf_stack);                                     \
     }                                                                   \
     leaf_stack_ptr = -1;                                               \
   } 

#define LEAF_STACK_PUSH(leaf) \
    {                                                            \
     if (leaf_stack_size <= leaf_stack_ptr) {                   \
       leaf_stack_size *= 2;                                    \
       leaf_stack = PIRAT_REALLOC_ARRAY (leaf_stack,             \
                                        leaf_stack_size,        \
                                        pirat_leaf_t *);        \
       PIRAT_CHECK_NEW (leaf_stack);                             \
     }                                                           \
     leaf_stack_ptr++;                                          \
     leaf_stack[leaf_stack_ptr] = leaf;                         \
   } 
#define LEAF_STACK_POP(leaf) \
    {                                                            \
     leaf = leaf_stack[leaf_stack_ptr];                         \
     leaf_stack_ptr--;                                          \
   } 
#define LEAF_STACK_EMPTY() (leaf_stack_ptr < 0) 

static void
pirat_descend_check_leaf (pirat_front_end_t *pfe,
                          pirat_leaf_t *root,
                          G3D_box_t *range
                          ) 
{
  pirat_leaf_t *cl; 
  
  LEAF_STACK_CLEAR ();
  LEAF_STACK_PUSH (root);

  do  {
    LEAF_STACK_POP (cl);
    if (cl->descend[0] == NULL) {
      if (G3D_BOXES_OVERLAP (pirat_range, cl->box)) {
        if (cl->pts) 
          pirat_check_leaf (cl);
      } 
    }  else  {
      switch (cl->split_across) {
      case X_SPLIT:
        if (G3D_RANGES_OVERLAP (pirat_range, cl->descend[0]->box, x)) {
          LEAF_STACK_PUSH (cl->descend[0]);
        } 
        if (G3D_RANGES_OVERLAP (pirat_range, cl->descend[1]->box, x)) {
          LEAF_STACK_PUSH (cl->descend[1]);
        } 
        break;
      case Y_SPLIT:
        if (G3D_RANGES_OVERLAP (pirat_range, cl->descend[0]->box, y)) {
          LEAF_STACK_PUSH (cl->descend[0]);
        } 
        if (G3D_RANGES_OVERLAP (pirat_range, cl->descend[1]->box, y)) {
          LEAF_STACK_PUSH (cl->descend[1]);
        } 
        break;
      default:
        if (G3D_RANGES_OVERLAP (pirat_range, cl->descend[0]->box, z)) {
          LEAF_STACK_PUSH (cl->descend[0]);
        } 
        if (G3D_RANGES_OVERLAP (pirat_range, cl->descend[1]->box, z)) {
          LEAF_STACK_PUSH (cl->descend[1]);
        } 
        break;
      } 
    } 
  }  while (!LEAF_STACK_EMPTY ());

  return;
} 
#else  /* recursive descent */

static void
pirat_descend_check_leaf (pirat_front_end_t *pfe,
                          pirat_leaf_t *leaf,
                          G3D_box_t *range
                          ) 
{
  if (leaf->descend[0] == NULL) {
    if (G3D_BOXES_OVERLAP (*range, leaf->box)) 
      if (leaf->pts) 
        pirat_check_leaf (pfe, leaf, range);
  }  else  {
    switch (leaf->split_across) {
    case X_SPLIT:
      if (G3D_RANGES_OVERLAP (*range, leaf->descend[0]->box, x)) {
        pirat_descend_check_leaf (pfe, leaf->descend[0], range);
      } 
      if (G3D_RANGES_OVERLAP (*range, leaf->descend[1]->box, x)) {
        pirat_descend_check_leaf (pfe, leaf->descend[1], range);
      } 
      break;
    case Y_SPLIT:
      if (G3D_RANGES_OVERLAP (*range, leaf->descend[0]->box, y)) {
        pirat_descend_check_leaf (pfe, leaf->descend[0], range);
      } 
      if (G3D_RANGES_OVERLAP (*range, leaf->descend[1]->box, y)) {
        pirat_descend_check_leaf (pfe, leaf->descend[1], range);
      } 
      break;
    default:
      if (G3D_RANGES_OVERLAP (*range, leaf->descend[0]->box, z)) {
        pirat_descend_check_leaf (pfe, leaf->descend[0], range);
      } 
      if (G3D_RANGES_OVERLAP (*range, leaf->descend[1]->box, z)) {
        pirat_descend_check_leaf (pfe, leaf->descend[1], range);
      } 
      break;
    } 
  } 
  return;
} 
#endif



static void
pirat_append_output (pirat_front_end_t *pfe, void *object);


static int
pirat_check_leaf (pirat_front_end_t *pfe, pirat_leaf_t *leaf,
                  G3D_box_t *range
                          ) 
{
  int c;

  c = 0;
  while (c < leaf->num_of_pts) {
    if (G3D_PT_INSIDE_BOX (*range, leaf->pts[c].loc)) {
      pirat_append_output (pfe, leaf->pts[c].object);
      if (pfe->find_max_num_objects <= pfe->num_of_points_found) {
        return FALSE;
      }
    } 
    c++;
  } 
  
  return TRUE; 
} 


static void
pirat_append_output (pirat_front_end_t *pfe, void *object) 
{
  pfe->num_of_points_found++;
  if (pfe->num_of_points_found > pfe->size_of_buffer) {
    pfe->size_of_buffer *= 2;
    pfe->entities_found
      = PIRAT_REALLOC_ARRAY (pfe->entities_found,
                             pfe->size_of_buffer, void *);
    PIRAT_CHECK_NEW (pfe->entities_found);
    
  } 
  pfe->entities_found[pfe->num_of_points_found - 1] = object;
} 


static int
pirat_extend (pirat_front_end_t *pfe, pirat_leaf_t *parent,
             pirat_leaf_t *leaf,
             void *object, G3D_loc_t *loc) ;
static int
pirat_split_leaf (pirat_front_end_t *pfe, pirat_leaf_t *leaf);
static pirat_leaf_t *
pirat_new_leaf (G3D_box_t *box, pirat_leaf_t *parent);


/*
  The function adds a point to the search tree.
 */
int
pirat_add_pt (pirat_front_end_t *pfe, void *object, G3D_loc_t *loc) 
{
  if (pfe->root == NULL) {
    G3D_box_t box;
    G3D_INIT_BOX (box, *loc);
    pfe->root = pirat_new_leaf (&box, NULL);
  } 

  pfe->num_of_inserted_points++;

  return pirat_extend (pfe, NULL, pfe->root, object, loc);
} 





static pirat_leaf_t *
pirat_new_leaf (G3D_box_t *box, pirat_leaf_t *parent) 
{
  pirat_leaf_t *l;

  l = PIRAT_NEW (pirat_leaf_t);
  PIRAT_CHECK_NEW (l);

  l->descend[0]   = NULL;
  l->descend[1]   = NULL;
  G3D_COPY_BOX (l->box, *box);
  l->num_of_pts   = 0;
  l->pts          = NULL;
  l->split_across = ' ';

  return l;
} 









static int
pirat_store (pirat_front_end_t *pfe,
             pirat_leaf_t *leaf, void *object, G3D_loc_t *loc) ;
static void
pirat_grow_self (pirat_front_end_t *pfe,
                 pirat_leaf_t *parent,
                 pirat_leaf_t *leaf,
                 G3D_loc_t *loc );



static int
pirat_extend (pirat_front_end_t *pfe, pirat_leaf_t *parent,
             pirat_leaf_t *leaf,
             void *object, G3D_loc_t *loc) 
{
  G3D_box_t b0, b1;

  pirat_grow_self (pfe, parent, leaf, loc);

 restart:
  if (leaf->descend[0] == NULL) { /* terminal leaf */
    if (G3D_PT_INSIDE_BOX_USTRICT (leaf->box, *loc)) {
      if (!pirat_store (pfe, leaf, object, loc)) {
        pirat_split_leaf (pfe, leaf);
        goto restart;
      } 
    } 
  }  else  {
    pirat_extend (pfe, leaf, leaf->descend[0], object, loc);
    pirat_extend (pfe, leaf, leaf->descend[1], object, loc);
  } 
  
  return TRUE;
} 




static int
pirat_store (pirat_front_end_t *pfe,
             pirat_leaf_t *leaf, void *object, G3D_loc_t *loc) 
{
  if (leaf->pts == NULL) {
    leaf->pts = pirat_new_pts_array (pfe);
    leaf->num_of_pts = 0;
  } 
  if (leaf->num_of_pts >= pfe->max_num_of_points_per_box) {
    return FALSE;
  }  else  {
    leaf->pts[leaf->num_of_pts].object = object;
    leaf->pts[leaf->num_of_pts].loc    = *loc;
    leaf->num_of_pts++;
    return TRUE;
  } 
} 



static int
pirat_split_leaf (pirat_front_end_t *pfe, pirat_leaf_t *leaf) 
{
  G3D_box_t b0, b1; 
  int i;

  leaf->split_across = split_dir (leaf);

  switch (leaf->split_across) {
  case X_SPLIT:
    G3D_SPLIT_ACROSS_DIR (b0, b1, leaf->box, x);
    break;
  case Y_SPLIT:
    G3D_SPLIT_ACROSS_DIR (b0, b1, leaf->box, y);
    break;
  default:
    G3D_SPLIT_ACROSS_DIR (b0, b1, leaf->box, z);
    break;
  } 

  leaf->descend[0] = pirat_new_leaf (&b0, leaf);
  leaf->descend[1] = pirat_new_leaf (&b1, leaf);

  leaf->descend[0]->pts = pirat_new_pts_array (pfe);
  leaf->descend[1]->pts = pirat_new_pts_array (pfe);

  FOR (i, leaf->num_of_pts) {
    if (G3D_PT_INSIDE_BOX_USTRICT (leaf->descend[0]->box,
                                       leaf->pts[i].loc)) {
      if (!pirat_store (pfe, leaf->descend[0],
                       leaf->pts[i].object,
                       &leaf->pts[i].loc)) {
        REPORT_ERROR ("?");
      } 
    }  else  {
      if (!pirat_store (pfe, leaf->descend[1],
                       leaf->pts[i].object,
                       &leaf->pts[i].loc)) {
        REPORT_ERROR ("?");
      } 
    } 
  } 

  pirat_free_pts_array (pfe, leaf->pts);
  leaf->pts = NULL; leaf->num_of_pts = 0;
  
  return TRUE;
} 



static void
pirat_grow_self (pirat_front_end_t *pfe,
                 pirat_leaf_t *parent,
                 pirat_leaf_t *leaf,
                 G3D_loc_t *loc) 
{
  G3D_box_t newbox;

  if (parent) {
    int is_lower = (parent->descend[0] == leaf);

    G3D_COPY_BOX (newbox, parent->box);
    switch (parent->split_across) {
    case X_SPLIT:
      if (is_lower) newbox.upper.x = leaf->box.upper.x;
      else          newbox.lower.x = leaf->box.lower.x;
      break;
    case Y_SPLIT:
      if (is_lower) newbox.upper.y = leaf->box.upper.y;
      else          newbox.lower.y = leaf->box.lower.y;
      break;
    default:
      if (is_lower) newbox.upper.z = leaf->box.upper.z;
      else          newbox.lower.z = leaf->box.lower.z;
      break;
    }  
    G3D_COPY_BOX (leaf->box, newbox);
  }  else  { /* no parent; no restriction growing */
    G3D_UPDT_BOX (leaf->box, *loc);
    G3D_INFLATE_BOX (leaf->box, pfe->inflate_tol);
  } 
} 



static int
pirat_descend_print_leaf (pirat_leaf_t *leaf, FILE *fp, int level);


int
pirat_print (pirat_front_end_t *pirat, FILE *fp) 
{
  int nt;
  
  fprintf (fp, "PiRaT structure:\n");
  fprintf (fp, "   max # of points per box %d\n\n",
	  pirat->max_num_of_points_per_box);
  nt = pirat_descend_print_leaf (pirat->root, fp, 1);
  fprintf (fp, "   total of inserted points %d\n",
	  pirat->num_of_inserted_points);
  fprintf (fp, "   total of actually stored points %d\n", nt);

  return TRUE;
} 


static int
pirat_print_leaf (pirat_leaf_t *leaf, FILE *fp, int level);

static int
pirat_descend_print_leaf (pirat_leaf_t *leaf, FILE *fp, int level) 
{
  int nt;
  
  nt = pirat_print_leaf (leaf, fp, level);
  if (leaf->descend[0]) 
    nt += pirat_descend_print_leaf (leaf->descend[0], fp, level+1);
  if (leaf->descend[1]) 
    nt += pirat_descend_print_leaf (leaf->descend[1], fp, level+1);

  return nt;
} 


static int
pirat_print_leaf (pirat_leaf_t *leaf, FILE *fp, int level) 
{
  int nt = 0;
  int i;
  
  fprintf (fp, "leaf %p\n", leaf);
  fprintf (fp, "   level   = %d\n", level);
  fprintf (fp, "   box     =  { {%g, %g, %g} ,  {%g, %g, %g} } \n",
	  leaf->box.lower.x,
	  leaf->box.lower.y,
	  leaf->box.lower.z,
	  leaf->box.upper.x,
	  leaf->box.upper.y,
	  leaf->box.upper.z);
  if (leaf->pts) {
    nt = leaf->num_of_pts;
    fprintf (fp, "   num_of_pts = %d\n", nt);
    FOR (i, leaf->num_of_pts) {
      fprintf (fp, "   pt[%d] =  {%g, %g, %g} \n",
              i, leaf->pts[i].loc.x, leaf->pts[i].loc.y, leaf->pts[i].loc.z);
    } 
  }  else
    fprintf (fp, "   descend =  {%p, %p} \n", leaf->descend[0], leaf->descend[1]);

  return nt;
} 


void *
pirat_first_in_range (pirat_front_end_t *pirat) 
{
  pirat->curr_entity = 0;
  return (pirat->num_of_points_found > 0
          ? pirat->entities_found[pirat->curr_entity]
          : NULL);
} 



void *
pirat_next_in_range (pirat_front_end_t *pirat) 
{
  pirat->curr_entity++;
  return (pirat->num_of_points_found > pirat->curr_entity
          ? pirat->entities_found[pirat->curr_entity]
          : NULL);
} 



static void
pirat_free_leaf (pirat_leaf_t *leaf) 
{
  if (leaf->pts) {
    PIRAT_FREE (leaf->pts);
  }  
  PIRAT_FREE (leaf);
} 



static void
pirat_descend_free_leaf (pirat_leaf_t *leaf)
{
  if (leaf->descend[0]) pirat_descend_free_leaf (leaf->descend[0]);
  if (leaf->descend[1]) pirat_descend_free_leaf (leaf->descend[1]);
  pirat_free_leaf (leaf);
} 



void
pirat_destroy_pirat (pirat_front_end_t *pirat) 
{
  if (pirat->entities_found) PIRAT_FREE (pirat->entities_found);
  if (pirat->free_pts_array) PIRAT_FREE (pirat->free_pts_array);
  pirat_descend_free_leaf (pirat->root);
} 



