#include <stdio.h>
#include <time.h>
#include "getoptP.h"
#include "pirat.h"
#include "piratutil.h"
#include "randP.h"
#include "timeP.h"
#include "prap3.h"


#define RUN_TEST 1
#define TIME_PRAP 1

#define NPTS   5000
#define NSRCHS NPTS

/*

5000 points, 5000 searches

prap: 16.86 sec (search), 0.02 secs (create)

pirat:   
n = max_num_of_points_per_box  
c = create time [secs]  
s = search time [secs]  
r = ratio prap / pirat (just search)
R = ratio prap / pirat (

    n            c             s            r             R
===========================================================================
    5           14.58        4.13       0.244958       1.10841
   10            8.25        3.27       0.19395        0.682464
   15            6.11        3.22       0.190985       0.552725
   20            4.9         2.81       0.166667       0.456754
   25            4.42        2.87       0.170225       0.431872
   30            4.09        2.73       0.161922       0.404028
   35            3.89        2.86       0.169632       0.400119        
   50            3.53        2.96       0.175563       0.384479
  100            3.08        3.51       0.208185       0.390403
*/
#define MAX_NUM_OF_POINTS_PER_BOX  30 

typedef struct pt {
  int        found_by_prap;
  int        found_by_pirat;
  int        found_by_brute;
  G3D_loc_t  loc;
}              pt;

static double xscalelo = -1;
static double xscalehi = 1;

static double yscalelo = 0.3;
static double yscalehi = 0.9;

static double zscalelo = 0; 
static double zscalehi = 1; 


static int num_of_points = NPTS;
static int num_of_searches = NSRCHS;
static pt *pts;
static double ustime, rstime, sstime;
static double uetime, retime, setime;
static double prap_time, pirat_time, prap_create, pirat_create;

static void 
generate_points(void);
static void
create_prap(void);
static void
create_pirat(void);
static void
test_pirat(void);
static void
time_prap(void);
static void
time_pirat(void);

int  
main(int argc, char **argv)
{
  fprintf(stderr, "generating points\n");
  generate_points();
  fprintf(stderr, "creating prap\n");
  create_prap();
  fprintf(stderr, "creating pirat\n");
  create_pirat();
#if RUN_TEST
  fprintf(stderr, "testing pirat\n");
  test_pirat();
#endif
#if TIME_PRAP
  fprintf(stderr, "timing prap\n");
  time_prap();
#else
  prap_time = 16.86;
#endif
  fprintf(stderr, "timing pirat\n");
  time_pirat();

  fprintf(stderr, "pirat_time / prap_time = %g\n", pirat_time / prap_time);
  fprintf(stderr, "(pirat_time + pirat_create) / (prap_time + prap_create) = %g\n",
          (pirat_time + pirat_create) / (prap_time + prap_create));

  return 0;
}


static void
generate_points(void)
{
  int i;

  pts = PIRAT_NEW_ARRAY(num_of_points, pt);
  PIRAT_CHECK_NEW(pts);

  FOR(i, num_of_points) {
    pts[i].loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    pts[i].loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    pts[i].loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  }
}


static char *ss;

static void
create_prap(void)
{
  int i;
  
  ckit_get_times(&rstime, &ustime, &sstime);
  ss = PRaP_new_search_structure(num_of_points / 10 + 1);
  FOR(i, num_of_points) {
    if (!PRaP_add_entity(ss, &pts[i], (double *)&pts[i].loc)) {
      REPORT_ERROR("failed PRaP_add_entity");
    }
  }
  ckit_get_times(&retime, &uetime, &setime);
  fprintf(stderr, "time for prap create = %g\n",
          prap_create = uetime - ustime);
}



static pirat_front_end_t *pirat = NULL;

static void
create_pirat(void)
{
  int i;

  ckit_get_times(&rstime, &ustime, &sstime);
  pirat = pirat_new_pirat(MAX_NUM_OF_POINTS_PER_BOX, 0.1);
  FOR(i, num_of_points) {
    if (!pirat_add_pt(pirat, &pts[i], &pts[i].loc)) {
      REPORT_ERROR("failed pirat_add_pt");
    }
#if 0
    fprintf(stderr, ">>>>>>>>>>>>>>>>>>>>>>\nadded pt # %d\n\n", i);
    pirat_print(pirat, stderr);
    fflush(stderr);
#endif
  }
  ckit_get_times(&retime, &uetime, &setime);
  fprintf(stderr, "time for pirat create = %g\n",
          pirat_create = uetime - ustime);
#if 0
  pirat_print(pirat, stderr);
  fflush(stderr);
#endif
}



static void
search_pirat(G3D_box_t *box);
static void
search_prap(G3D_box_t *box);
static void
search_brute(G3D_box_t *box);
static void
compare(G3D_box_t *box);



static void
test_pirat(void)
{
  int i, j;
  G3D_loc_t p;
  G3D_box_t box;

  FOR(i, num_of_searches) {
    FOR(j, num_of_points) {
      pts[j].found_by_prap = 0;
      pts[j].found_by_pirat = 0;
      pts[j].found_by_brute = 0;
    }
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_INIT_BOX(box, p);
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_UPDT_BOX(box, p);
    search_prap(&box);
    search_pirat(&box);
    search_brute(&box);
    compare(&box);
    if (i / 1000 * 1000 == i)
      fprintf(stderr, " done %d of %d\n", i, num_of_searches);
  }
}


static void
dump_data(G3D_box_t *box);

static void
compare(G3D_box_t *box)
{
  int i;

  FOR(i, num_of_points) {
    if (pts[i].found_by_prap != pts[i].found_by_pirat) {
      if (pts[i].found_by_pirat != pts[i].found_by_brute) {
        fprintf(stderr, "box = {{%g, %g, %g}, {%g, %g, %g}}\n",
                box->lower.x, box->lower.y, box->lower.z,
                box->upper.x, box->upper.y, box->upper.z);
        fprintf(stderr, "pt  = {%g, %g, %g}\n",
                pts[i].loc.x, pts[i].loc.y, pts[i].loc.z);
        fprintf(stderr, "status: prap = %d, brute = %d, pirat = %d\n",
                pts[i].found_by_prap,
                pts[i].found_by_brute,
                pts[i].found_by_pirat);
        pirat_search_range(pirat, box);
        dump_data(box);
        REPORT_ERROR("prap and pirat searches do not agree");
      }
    }
  }
}


static void
dump_data(G3D_box_t *box)
{
  int i;

  fprintf(stderr, "static G3D_box_t box = {\n");
  fprintf(stderr, "  {{%g,%g,%g}, {%g,%g,%g}\n",
          box->lower.x, box->lower.y, box->lower.z,
          box->upper.x, box->upper.y, box->upper.z);
  fprintf(stderr, "};\n");
  
  fprintf(stderr, "static pt points[] = {\n");
  FOR(i, num_of_points) {
    fprintf(stderr, "   {0,0,0,{%g,%g,%g}},\n",
            pts[i].loc.x, pts[i].loc.y, pts[i].loc.z);
  }
  fprintf(stderr, "};\n");
}


static void
search_brute(G3D_box_t *box)
{
  int i;

  FOR(i, num_of_points) {
    if (G3D_PT_INSIDE_BOX(*box, pts[i].loc)) {
      pts[i].found_by_brute = 1;
    } else {
      pts[i].found_by_brute = 0;
    }
  }
}


static void
search_prap(G3D_box_t *box)
{
  PRaP_range_t range;
  pt *p;
  
  range.lower[0] = box->lower.x;
  range.lower[1] = box->lower.y;
  range.lower[2] = box->lower.z;
  range.upper[0] = box->upper.x;
  range.upper[1] = box->upper.y;
  range.upper[2] = box->upper.z;
  if (!PRaP_search_range(ss, &range)) {
    REPORT_ERROR("error in PRaP_search_range");
  }
  p = PRaP_first_entity_in_range(ss);
  while (p != NULL) {
    p->found_by_prap++;
    p = PRaP_next_entity_in_range(ss);
  }
}



static void
search_pirat(G3D_box_t *box)
{
  pt *p;
  int i;
  
  pirat_search_range(pirat, box);

  FOR(i, pirat->num_of_points_found) {
    p = pirat->entities_found[i];
    p->found_by_pirat++;
  }
}



static void
time_pirat(void)
{
  int i;
  G3D_loc_t p;
  G3D_box_t box;

  ckit_get_times(&rstime, &ustime, &sstime);
  FOR(i, num_of_searches) {
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_INIT_BOX(box, p);
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_UPDT_BOX(box, p);
    search_pirat(&box);
  }
  ckit_get_times(&retime, &uetime, &setime);
  fprintf(stderr, "time for pirat = %g\n", pirat_time = uetime - ustime);
}



static void
time_prap(void)
{
  int i;
  G3D_loc_t p;
  G3D_box_t box;

  ckit_get_times(&rstime, &ustime, &sstime);
  FOR(i, num_of_searches) {
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_INIT_BOX(box, p);
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_UPDT_BOX(box, p);
    search_prap(&box);
  }
  ckit_get_times(&retime, &uetime, &setime);
  fprintf(stderr, "time for prap = %g\n", prap_time = uetime - ustime);
}

